@@ -22,24 +22,30 @@
 #include "config.h"
 #endif
 
+#include "videostream.h"
+
 #include <string.h>
 
 #include <telepathy-glib/errors.h>
+#include <telepathy-glib/dbus.h>
+#include <telepathy-glib/proxy-subclass.h>
 
 #include <gst/interfaces/xoverlay.h>
+#include <gst/farsight/fs-element-added-notifier.h>
 
-#include "videostream.h"
 #include "tp-stream-engine.h"
 #include "tp-stream-engine-signals-marshal.h"
 #include "util.h"
 
+#include "api/api.h"
+
 G_DEFINE_TYPE (TpStreamEngineVideoStream, tp_stream_engine_video_stream,
     TP_STREAM_ENGINE_TYPE_VIDEO_SINK);
 
-#define DEBUG(stream, format, ...) \
-  g_debug ("stream %d (video) %s: " format, \
-      ((TfStream *) stream)->stream_id,     \
-      G_STRFUNC,                            \
+#define DEBUG(self, format, ...)                \
+  g_debug ("stream %d (video) %s: " format,     \
+      tf_stream_get_id ((self)->priv->stream),  \
+      G_STRFUNC,                                \
       ##__VA_ARGS__)
 
 struct _TpStreamEngineVideoStreamPrivate
@@ -56,7 +62,11 @@ struct _TpStreamEngineVideoStreamPrivate
 
   GstElement *sink;
 
-  GstElement *queue;
+  gboolean manual_keyframes;
+  volatile gint bitrate;
+  volatile gint mtu;
+
+  FsElementAddedNotifier *notifier;
 
   GMutex *mutex;
 
@@ -72,6 +82,8 @@ struct _TpStreamEngineVideoStreamPrivate
 enum
 {
   RECEIVING,
+  RESOLUTION_CHANGED,
+  SINK_ADDED_SIGNAL,
   SIGNAL_COUNT
 };
 
@@ -107,6 +119,11 @@ static void tp_stream_engine_video_stream_get_property  (GObject *object,
     GValue *value,
     GParamSpec *pspec);
 
+static void _element_added (FsElementAddedNotifier *notifier,
+    GstBin *conference,
+    GstElement *element,
+    TpStreamEngineVideoStream *self);
+
 
 static void
 tp_stream_engine_video_stream_init (TpStreamEngineVideoStream *self)
@@ -114,9 +131,17 @@ tp_stream_engine_video_stream_init (TpStreamEngineVideoStream *self)
   TpStreamEngineVideoStreamPrivate *priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
       TP_STREAM_ENGINE_TYPE_VIDEO_STREAM, TpStreamEngineVideoStreamPrivate);
 
+  self->width = G_MAXUINT;
+  self->height = G_MAXUINT;
+
   self->priv = priv;
 
   self->priv->mutex = g_mutex_new ();
+
+  self->priv->notifier = fs_element_added_notifier_new ();
+
+  g_signal_connect (self->priv->notifier, "element-added",
+      G_CALLBACK (_element_added), self);
 }
 
 
@@ -200,6 +225,266 @@ receiving_data (GstPad *pad, GstMiniObject *obj, gpointer user_data)
   return TRUE;
 }
 
+static void
+request_keyframe (TpProxy *proxy, gpointer user_data, GObject *weak_object)
+{
+  TpStreamEngineVideoStream *self = TP_STREAM_ENGINE_VIDEO_STREAM (weak_object);
+
+  DEBUG (self, "Keyframe requested by connection manager");
+
+  if (self->priv->pad)
+    gst_pad_push_event (self->priv->pad,
+        gst_event_new_custom (GST_EVENT_CUSTOM_DOWNSTREAM,
+            gst_structure_new ("GstForceKeyUnit",
+                "all-headers", G_TYPE_BOOLEAN, TRUE,
+                NULL)));
+}
+
+static gboolean
+is_in_send_codecbin (GstElement *element, GstElement *ultimate_parent,
+    guint session_id)
+{
+  gint tmp_session_id, tmp_codec_id;
+  gchar *name;
+  gboolean valid = FALSE;
+  GstElement *parent;
+  gboolean rv;
+
+  if (element == ultimate_parent)
+    return FALSE;
+
+  name = gst_element_get_name (element);
+  if (name && sscanf (name, "send_%d_%d", &tmp_session_id, &tmp_codec_id) == 2)
+    valid = TRUE;
+  g_free (name);
+
+  if (valid && session_id == tmp_session_id)
+    return TRUE;
+
+  parent = (GstElement*) gst_element_get_parent (element);
+  if (!parent)
+    return FALSE;
+  rv = is_in_send_codecbin (parent, ultimate_parent, session_id);
+
+  gst_object_unref (parent);
+
+  return rv;
+}
+
+static void
+_element_added (FsElementAddedNotifier *notifier,
+    GstBin *bin, GstElement *element, TpStreamEngineVideoStream *self)
+{
+  gint bitrate = g_atomic_int_get (&self->priv->bitrate);
+  gint mtu = g_atomic_int_get (&self->priv->mtu);
+
+  if (G_UNLIKELY (
+          (bitrate && g_object_has_property (G_OBJECT (element), "bitrate")) ||
+          (mtu && g_object_has_property (G_OBJECT (element), "mtu")) ||
+          (self->priv->manual_keyframes &&
+              g_object_has_property (G_OBJECT (element), "keyframe-interval"))))
+    {
+      FsSession *fssession;
+      GstElement *conf;
+      guint id;
+
+      g_object_get (self->priv->stream, "farsight-session", &fssession,
+          "farsight-conference", &conf, NULL);
+
+      g_object_get (fssession, "id", &id, NULL);
+
+      if (is_in_send_codecbin (element, GST_ELEMENT (conf), id))
+        {
+          if (bitrate && g_object_has_property (G_OBJECT (element), "bitrate"))
+            {
+              g_object_set (element, "bitrate", bitrate, NULL);
+              DEBUG (self, "Set bitrate %u on encoder", bitrate);
+            }
+          if (mtu && g_object_has_property (G_OBJECT (element), "mtu"))
+            {
+              g_object_set (element, "mtu", mtu, NULL);
+              DEBUG (self, "Set MTU %u on payloader", mtu);
+            }
+          if (self->priv->manual_keyframes &&
+              g_object_has_property (G_OBJECT (element), "keyframe-interval"))
+            {
+              g_object_set (element, "keyframe-interval", 600, NULL);
+              DEBUG (self, "Disabled automatic keyframes");
+            }
+        }
+
+      g_object_unref (fssession);
+    }
+}
+
+static void
+tp_stream_engine_video_stream_update_element_properties (
+    TpStreamEngineVideoStream *self)
+{
+  GstBin *conf;
+
+  g_object_get (self->priv->stream, "farsight-conference", &conf, NULL);
+
+  fs_element_added_notifier_remove (self->priv->notifier, conf);
+
+  if (g_atomic_int_get (&self->priv->bitrate) ||
+      g_atomic_int_get (&self->priv->mtu) ||
+      self->priv->manual_keyframes)
+    fs_element_added_notifier_add (self->priv->notifier, conf);
+
+  gst_object_unref (conf);
+}
+
+static void
+tp_stream_engine_video_stream_set_bitrate (TpStreamEngineVideoStream *self,
+    guint bitrate)
+{
+  if (g_atomic_int_get (&self->priv->bitrate) == bitrate)
+    return;
+
+  g_atomic_int_set (&self->priv->bitrate, bitrate);
+
+  tp_stream_engine_video_stream_update_element_properties (self);
+}
+
+
+static void
+bitrate_changed (TpProxy *proxy, guint bitrate, gpointer user_data,
+    GObject *weak_object)
+{
+  TpStreamEngineVideoStream *self = TP_STREAM_ENGINE_VIDEO_STREAM (weak_object);
+
+  DEBUG (self, "BitrateChanged: %u", bitrate);
+
+  tp_stream_engine_video_stream_set_bitrate (self, bitrate);
+}
+
+static void
+tp_stream_engine_video_stream_set_mtu (TpStreamEngineVideoStream *self,
+    guint mtu)
+{
+  if (g_atomic_int_get (&self->priv->mtu) == mtu)
+    return;
+
+  g_atomic_int_set (&self->priv->mtu, mtu);
+
+  tp_stream_engine_video_stream_update_element_properties (self);
+}
+
+
+static void
+mtu_changed (TpProxy *proxy, guint mtu, gpointer user_data,
+    GObject *weak_object)
+{
+  TpStreamEngineVideoStream *self = TP_STREAM_ENGINE_VIDEO_STREAM (weak_object);
+
+  DEBUG (self, "MTUChanged: %u", mtu);
+
+  tp_stream_engine_video_stream_set_mtu (self, mtu);
+}
+
+static void
+tp_stream_engine_video_stream_set_manual_keyframes (
+    TpStreamEngineVideoStream *self)
+{
+  self->priv->manual_keyframes = TRUE;
+
+  tp_stream_engine_video_stream_update_element_properties (self);
+}
+
+
+static void
+video_resolution_changed_real (TpStreamEngineVideoStream *self,
+    guint width, guint height)
+{
+  DEBUG (self, "ResolutionChanged: %u x %u", width, height);
+
+  if (width != self->width || height != self->height)
+    {
+      self->width = width;
+      self->height = height;
+      g_signal_emit (self, signals[RESOLUTION_CHANGED], 0);
+    }
+}
+
+static void
+video_resolution_changed (TpProxy *proxy, const GValueArray *arg_resolution,
+    gpointer user_data, GObject *weak_object)
+{
+  TpStreamEngineVideoStream *self = TP_STREAM_ENGINE_VIDEO_STREAM (weak_object);
+  GValue *val_width, *val_height;
+
+  if (arg_resolution->n_values != 2)
+    {
+      DEBUG (self,
+          "VideoResolutionChanged argument does not have two members");
+      return;
+    }
+
+  val_width = g_value_array_get_nth (arg_resolution, 0);
+  val_height = g_value_array_get_nth (arg_resolution, 1);
+
+  if (!G_VALUE_HOLDS_UINT (val_width) || !G_VALUE_HOLDS_UINT (val_height))
+    {
+      DEBUG (self, "VideoResolutionChanged arguments are not uints");
+      return;
+    }
+
+  video_resolution_changed_real (self, g_value_get_uint (val_width),
+      g_value_get_uint (val_height));
+
+}
+static void
+videocontrol_get_all_properties_cb (TpProxy *proxy, GHashTable *out_properties,
+    const GError *error, gpointer user_data, GObject *weak_object)
+{
+  TpStreamEngineVideoStream *self = TP_STREAM_ENGINE_VIDEO_STREAM (weak_object);
+  guint32 bitrate;
+  guint32 mtu;
+  gboolean manual_keyframes;
+  gboolean valid;
+  GValueArray *array;
+
+  if (error)
+    return;
+
+  bitrate = tp_asv_get_uint32 (out_properties, "Bitrate", &valid);
+  if (valid)
+    {
+      DEBUG (self, "Got initial bitrate: %u", bitrate);
+      tp_stream_engine_video_stream_set_bitrate (self, bitrate);
+    }
+
+  mtu = tp_asv_get_uint32 (out_properties, "MTU", &valid);
+  if (valid)
+    {
+      DEBUG (self, "Got initial mtu: %u", mtu);
+      tp_stream_engine_video_stream_set_mtu (self, mtu);
+    }
+
+  array = tp_asv_get_boxed (out_properties, "VideoResolution",
+      STREAM_ENGINE_STRUCT_TYPE_VIDEO_RESOLUTION);
+  if (array)
+    {
+      int width = g_value_get_uint (g_value_array_get_nth (array, 0));
+      int height = g_value_get_uint (g_value_array_get_nth (array, 1));
+
+      DEBUG (self, "Got initial resolution: %u x %u", width, height);
+
+      video_resolution_changed_real (self, width, height);
+    }
+
+  manual_keyframes = tp_asv_get_boolean (out_properties, "ManualKeyframes", &valid);
+  if (valid)
+    {
+      DEBUG (self, "Keyframes will only come when manually requested");
+      self->priv->manual_keyframes = TRUE;
+      tp_stream_engine_video_stream_update_element_properties (self);
+    }
+
+}
+
+
 static GObject *
 tp_stream_engine_video_stream_constructor (GType type,
     guint n_props,
@@ -207,41 +492,39 @@ tp_stream_engine_video_stream_constructor (GType type,
 {
   GObject *obj;
   TpStreamEngineVideoStream *self = NULL;
-  GstPad *srcpad;
   GstPad *sinkpad;
+  TpProxy *proxy;
+  GError *error = NULL;
 
   obj = G_OBJECT_CLASS (tp_stream_engine_video_stream_parent_class)->constructor (type, n_props, props);
 
   self = (TpStreamEngineVideoStream *) obj;
 
-  self->priv->queue = gst_element_factory_make ("queue", NULL);
+  g_object_get (self->priv->stream, "proxy", &proxy, NULL);
 
-  if (!self->priv->queue)
+  tp_proxy_add_interface_by_id (proxy,
+      STREAM_ENGINE_IFACE_QUARK_STREAMHANDLER_INTERFACE_VIDEO_CONTROL);
+  stream_engine_cli_streamhandler_interface_video_control_connect_to_request_keyframe (proxy, request_keyframe, NULL, NULL, G_OBJECT (self), &error);
+  g_assert (error == NULL);
+  stream_engine_cli_streamhandler_interface_video_control_connect_to_bitrate_changed (proxy, bitrate_changed, NULL, NULL, G_OBJECT (self), &error);
+  g_assert (error == NULL);
+  stream_engine_cli_streamhandler_interface_video_control_connect_to_mtu_changed (proxy, mtu_changed, NULL, NULL, G_OBJECT (self), &error);
+  g_assert (error == NULL);
+  stream_engine_cli_streamhandler_interface_video_control_connect_to_video_resolution_changed (proxy, video_resolution_changed, NULL, NULL, G_OBJECT (self), &error);
+  g_assert (error == NULL);
+  tp_cli_dbus_properties_call_get_all (proxy, -1,
+      STREAM_ENGINE_IFACE_STREAMHANDLER_INTERFACE_VIDEO_CONTROL,
+      videocontrol_get_all_properties_cb, NULL, NULL, G_OBJECT (self));
+  g_object_unref (proxy);
+
+  g_object_get (self->priv->stream, "sink-pad", &sinkpad, NULL);
+
+  if (!sinkpad)
     {
-      g_warning ("Could not make queue element");
+      g_warning ("Could not get stream's sinkpad");
       return obj;
     }
 
-  g_object_set (self->priv->queue,
-      "leaky", 2,
-      "max-size-buffers", 1,
-      NULL);
-
-  if (!gst_bin_add (GST_BIN (self->priv->bin), self->priv->queue))
-    {
-      g_warning ("Could not add quue to bin");
-      return obj;
-    }
-
-  if (gst_element_set_state (self->priv->queue, GST_STATE_PLAYING) ==
-      GST_STATE_CHANGE_FAILURE)
-    {
-      g_warning ("Could not start queue");
-      return obj;
-    }
-
-  sinkpad = gst_element_get_static_pad (self->priv->queue, "sink");
-
   if (GST_PAD_LINK_FAILED (gst_pad_link (self->priv->pad, sinkpad)))
     {
       g_warning ("Could not link sink to queue");
@@ -251,37 +534,9 @@ tp_stream_engine_video_stream_constructor (GType type,
 
   gst_object_unref (sinkpad);
 
-  g_object_get (self->priv->stream, "sink-pad", &sinkpad, NULL);
-
-  if (!sinkpad)
-    {
-      g_warning ("Could not get stream's sinkpad");
-      return obj;
-    }
-
-  srcpad = gst_element_get_static_pad (self->priv->queue, "src");
-
-  if (!sinkpad)
-    {
-      g_warning ("Could not get queue's srcpad");
-      gst_object_unref (sinkpad);
-      return obj;
-    }
-
-  if (GST_PAD_LINK_FAILED (gst_pad_link (srcpad, sinkpad)))
-    {
-      gst_object_unref (srcpad);
-      gst_object_unref (sinkpad);
-      g_warning ("Could not link sink to queue");
-      return obj;
-    }
-
-  gst_object_unref (srcpad);
-  gst_object_unref (sinkpad);
-
-
-  self->priv->src_pad_added_handler_id = g_signal_connect (self->priv->stream,
-      "src-pad-added", G_CALLBACK (src_pad_added_cb), self);
+  self->priv->src_pad_added_handler_id = g_signal_connect_object (
+      self->priv->stream, "src-pad-added", G_CALLBACK (src_pad_added_cb), self,
+      0);
 
   return obj;
 }
@@ -299,6 +554,12 @@ tp_stream_engine_video_stream_dispose (GObject *object)
       self->priv->src_pad_added_handler_id = 0;
     }
 
+  if (self->priv->notifier)
+    {
+      g_object_unref (self->priv->notifier);
+      self->priv->notifier = NULL;
+    }
+
   g_mutex_lock (self->priv->mutex);
   if (self->priv->receiving_probe_id)
     {
@@ -318,14 +579,6 @@ tp_stream_engine_video_stream_dispose (GObject *object)
     }
   g_mutex_unlock (self->priv->mutex);
 
-  if (self->priv->queue)
-    {
-      gst_element_set_locked_state (self->priv->queue, TRUE);
-      gst_element_set_state (self->priv->queue, GST_STATE_NULL);
-      gst_bin_remove (GST_BIN (self->priv->bin), self->priv->queue);
-      self->priv->queue = NULL;
-    }
-
   if (self->priv->sink)
     {
       gst_element_set_locked_state (self->priv->sink, TRUE);
@@ -460,6 +713,23 @@ tp_stream_engine_video_stream_class_init (TpStreamEngineVideoStreamClass *klass)
           NULL, NULL,
           g_cclosure_marshal_VOID__VOID,
           G_TYPE_NONE, 0);
+
+  signals[RESOLUTION_CHANGED] =
+      g_signal_new ("resolution-changed",
+          G_OBJECT_CLASS_TYPE (klass),
+          G_SIGNAL_RUN_LAST,
+          0,
+          NULL, NULL,
+          g_cclosure_marshal_VOID__VOID,
+          G_TYPE_NONE, 0);
+
+  signals[SINK_ADDED_SIGNAL] = g_signal_new ("sink-added",
+      G_OBJECT_CLASS_TYPE (klass),
+      G_SIGNAL_RUN_LAST,
+      0,
+      NULL, NULL,
+      g_cclosure_marshal_VOID__OBJECT,
+      G_TYPE_NONE, 1, G_TYPE_OBJECT);
 }
 
 
@@ -468,10 +738,9 @@ src_pad_added_idle_error (gpointer user_data)
 {
   TpStreamEngineVideoStream *self = TP_STREAM_ENGINE_VIDEO_STREAM (user_data);
 
-  tf_stream_error (self->priv->stream, 0,
+  tf_stream_error (self->priv->stream, TP_MEDIA_STREAM_ERROR_MEDIA_ERROR,
       "Error setting up video reception");
 
-
   g_mutex_lock (self->priv->mutex);
   self->priv->error_idle_id = 0;
   g_mutex_unlock (self->priv->mutex);
@@ -522,6 +791,7 @@ src_pad_added_cb (TfStream *stream, GstPad *pad, FsCodec *codec,
           g_warning ("Could not start sink");
           return;
         }
+
     }
 
   funnel = gst_bin_get_by_name (GST_BIN (self->priv->sink), "funnel");
@@ -535,6 +805,7 @@ src_pad_added_cb (TfStream *stream, GstPad *pad, FsCodec *codec,
       goto error;
     }
 
+  g_signal_emit (self, signals[SINK_ADDED_SIGNAL], 0, funnel);
 
   sinkpad = gst_element_get_request_pad (funnel, "sink%d");
   if (!sinkpad)
